// mod db;
mod gpo;
mod serve;

use std::path::PathBuf;

use clap::{Args, Parser, Subcommand, ValueEnum};
use figment::{
    providers::{Env, Format, Serialized, Toml},
    Figment,
};
use serde::Serialize;

use crate::config::LogLevel;

/// Otter is a lightweight implementation of the Gpodder API, designed to be used for small
/// personal deployments.
#[derive(Parser)]
pub struct Cli {
    #[command(flatten)]
    pub config: ClapConfig,

    #[command(subcommand)]
    pub cmd: Command,
}

#[derive(Serialize, Args, Clone, Debug)]
pub struct ClapConfig {
    #[arg(
        short,
        long = "config",
        env = "OTTER_CONFIG_FILE",
        value_name = "CONFIG_FILE",
        global = true
    )]
    config_file: Option<PathBuf>,

    #[command(flatten)]
    net: NetConfig,

    #[serde(skip_serializing_if = "Option::is_none")]
    #[arg(long = "data", value_name = "DATA_DIR", global = true)]
    data_dir: Option<PathBuf>,

    #[serde(skip_serializing_if = "Option::is_none")]
    #[arg(short, long = "log", value_name = "LOG_LEVEL", global = true)]
    log_level: Option<LogLevel>,
}

#[derive(ValueEnum, Serialize, Clone, Debug)]
#[serde(rename_all = "lowercase")]
pub enum NetConfigType {
    Tcp,
    Unix,
}

#[derive(Serialize, Clone, Args, Debug)]
pub struct NetConfig {
    #[serde(skip_serializing_if = "Option::is_none")]
    #[arg(long = "net", value_name = "NET", global = true)]
    r#type: Option<NetConfigType>,

    #[serde(skip_serializing_if = "Option::is_none")]
    #[arg(long, value_name = "DOMAIN", global = true)]
    domain: Option<String>,

    #[serde(skip_serializing_if = "Option::is_none")]
    #[arg(long, value_name = "PORT", global = true)]
    port: Option<u16>,

    #[serde(skip_serializing_if = "Option::is_none")]
    #[serde(rename = "path")]
    #[arg(long, value_name = "SOCKET", global = true)]
    socket: Option<PathBuf>,
}

#[derive(Subcommand)]
pub enum Command {
    Serve,
    // #[command(subcommand)]
    // Db(db::DbCommand),
    /// Perform operations on the database through the Gpodder abstraction, allowing operations
    /// identical to the ones performed by the API.
    #[command(subcommand)]
    Gpo(gpo::Command),
}

pub struct CliError {
    msg: Option<String>,
    err: Box<dyn std::error::Error>,
}

impl<E: std::error::Error + 'static> From<E> for CliError {
    fn from(error: E) -> Self {
        Self {
            msg: None,
            err: Box::new(error),
        }
    }
}

impl std::fmt::Display for CliError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if let Some(msg) = &self.msg {
            write!(f, "{msg}: ")?;
        }

        write!(f, "{}", self.err)
    }
}

impl CliError {
    pub fn msg(mut self, msg: impl AsRef<str>) -> Self {
        self.msg = Some(String::from(msg.as_ref()));

        self
    }
}

impl Cli {
    pub fn run(&self) -> Result<(), CliError> {
        let mut figment =
            Figment::new().merge(Serialized::defaults(crate::config::Config::default()));

        if let Some(config_path) = &self.config.config_file {
            figment = figment.merge(Toml::file(config_path));
        }

        let config: crate::config::Config = figment
            .merge(Env::prefixed("OTTER_"))
            .merge(Serialized::defaults(self.config.clone()))
            .extract()
            .map_err(|err| CliError::from(err).msg("failed to read config"))?;

        match &self.cmd {
            Command::Serve => serve::serve(&config),
            // Command::Db(cmd) => cmd.run(&config),
            Command::Gpo(cmd) => cmd.run(&config),
        }
    }
}
