use std::{sync::Arc, time::Duration};

use crate::{config::NetConfig, server};

use super::CliError;

pub fn serve(config: &crate::config::Config) -> Result<(), CliError> {
    tracing_subscriber::fmt()
        .with_max_level(tracing::Level::from(config.log_level))
        .init();

    tracing::info!("Initializing database and running migrations");

    let store =
        gpodder_sqlite::SqliteRepository::from_path(config.data_dir.join(crate::DB_FILENAME))
            .map_err(|err| CliError::from(err).msg("failed to initialize Sqlite repository"))?;

    // SAFETY this runs on statically embedded templates, and should therefore never fail
    let tera = crate::web::initialize_tera().unwrap();

    let store = gpodder::GpodderRepository::new(store);

    let ctx = server::Context {
        store,
        tera: Arc::new(tera),
    };
    let app = server::app(ctx.clone());

    let rt = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()
        .map_err(|err| CliError::from(err).msg("failed to initialize Tokio runtime"))?;

    // Spawn the session cleanup background task
    let session_removal_duration = Duration::from_secs(config.session_cleanup_interval);

    tracing::info!("Starting session cleanup background task");

    rt.spawn(async move {
        let mut interval = tokio::time::interval(session_removal_duration);

        loop {
            interval.tick().await;

            tracing::info!("Performing session cleanup");

            match ctx.store.remove_old_sessions() {
                Ok(n) => {
                    tracing::info!("Removed {} old sessions", n);
                }
                Err(err) => {
                    tracing::error!("Error occured during session cleanup: {}", err);
                }
            }
        }
    });

    tracing::info!("Initializing server");

    match &config.net {
        NetConfig::Tcp { domain, port } => {
            let address = format!("{}:{}", domain, port);
            tracing::info!("Listening on TCP address {address}");

            rt.block_on(async {
                let listener = tokio::net::TcpListener::bind(address)
                    .await
                    .map_err(|err| CliError::from(err).msg("failed to bind TCP listener"))?;

                axum::serve(listener, app.into_make_service())
                    .await
                    .map_err(|err| {
                        CliError::from(err).msg("error occured while running Axum server")
                    })
            })?;
        }
        NetConfig::Unix { path } => {
            // Try to remove the socket file first if it exists
            let _ = std::fs::remove_file(&path);

            if let Some(parent) = path.parent() {
                std::fs::create_dir_all(parent).unwrap();
            }

            tracing::info!("Listening on Unix socket {:?}", path);

            rt.block_on(async {
                let listener = tokio::net::UnixListener::bind(path)
                    .map_err(|err| CliError::from(err).msg("failed to bind Unix listener"))?;

                axum::serve(listener, app.into_make_service())
                    .await
                    .map_err(|err| {
                        CliError::from(err).msg("error occured while running Axum server")
                    })
            })?;
        }
    }

    Ok(())
}
